from .simplified_http_client import Client, Response
from .exceptions import *

class Interface(object):
    def __init__(self, host, auth):
        """
        Construct the basic Interface object for HTTP calls. 
        The HTTP client is being set up during initialization,
        changes in runtime will not affect its behaviour.
        :param host: base URL for API calls
        :type host: string
        :param auth: the authorization header
        :type auth: string
        """
        from . import __version__
        self.auth = auth
        self.host = host
        self.version = __version__
        self.useragent = 'loyalicos/{};python'.format(self.version)
        self.method = None
        self.path = None
        self.data = None
        self.json = None
        self.params = None
        self.timeout = None

        self.client = Client(
            host=self.host,
            request_headers=self._default_headers,
            version=1)

    @property
    def _default_headers(self):
        """Set the default header for a Loyalicos API call"""
        headers = {
            "Authorization": self.auth,
            "User-Agent": self.useragent,
            "Accept": 'application/json'
        }
        return headers

    def reset_request_headers(self):
        self.client.request_headers = self._default_headers

    def update_headers(self, new_headers, replace=False):
        if replace == True:
            self.client.request_headers = new_headers
        else:
            current_headers = self.client.request_headers
            current_headers.update(new_headers)
            self.client.request_headers = current_headers

    def send_request(self, method=None, path=None, json=None, data=None, params=None, timeout=None):
        self.method = method or self.method
        self.path = path or self.path
        self.data = data or self.data
        self.json = json or self.json
        self.params = params or self.params
        self.timeout = timeout or self.timeout
        if self.method == None:
            raise RequestNotReadyError
        self.response = Response(self.client.make_request(self.method, self.path, json=self.json, data=self.data, params=self.params, timeout=self.timeout))


class Error(Exception):
    """Base class for other exceptions"""
    pass


class NoCredentialsFoundError(Exception):
    """Raised when no API credentials are found"""

    def __init__(self, message="Nor API Key nor client/secret provided to API Constructor"):
        self.message = message
        super().__init__(self.message)

    # def __str__(self):
    #     return f'{self.salary} -> {self.message}'


class MemberNotFoundError(Exception):
    """Raised when no Member not found"""

    def __init__(self, message="Member not found"):
        self.message = message
        super().__init__(self.message)

    # def __str__(self):
    #     return f'{self.salary} -> {self.message}'


class DuplicateKeyForMemberError(Exception):
    """Raised when the same key is created twice"""

    def __init__(self, message="Member duplicate key"):
        self.message = message
        super().__init__(self.message)

    # def __str__(self):
    #     return f'{self.salary} -> {self.message}'


class RequestNotReadyError(Exception):
    """Raised when an imcomplete request is sent"""

    def __init__(self, message="Method None is nos supported"):
        self.message = message
        super().__init__(self.message)

    # def __str__(self):
    #     return f'{self.salary} -> {self.message}'

class HTTPRequestError(Exception):
    """Raised when a request fails"""
    pass


    status_code_dict= {
        100: "continue",
        101: "switching_protocols",
        102: "processing",
        103: "checkpoint",
        122: "uri_too_long, request_uri_too_long",
        200: "ok, okay, all_ok, all_okay, all_good, \o/, ✓",
        201: "created",
        202: "accepted",
        203: "non_authoritative_info, non_authoritative_information",
        204: "no_content",
        205: "reset_content, reset",
        206: "partial_content, partial",
        207: "multi_status, multiple_status, multi_stati, multiple_stati",
        208: "already_reported",
        226: "im_used",
        300: "multiple_choices",
        301: "moved_permanently, moved",
        302: "found",
        303: "see_other, other",
        304: "not_modified",
        305: "use_proxy",
        306: "switch_proxy",
        307: "temporary_redirect, temporary_moved, temporary",
        308: "permanent_redirect, resume_incomplete, resume",
        400: "bad_request, bad",
        401: "unauthorized",
        402: "payment_required, payment",
        403: "forbidden",
        404: "not_found, -o-",
        405: "method_not_allowed, not_allowed",
        406: "not_acceptable",
        407: "proxy_authentication_required, proxy_auth, proxy_authentication",
        408: "request_timeout, timeout",
        409: "conflict",
        410: "gone",
        411: "length_required",
        412: "precondition_failed, precondition",
        413: "request_entity_too_large",
        414: "request_uri_too_large",
        415: "unsupported_media_type, unsupported_media, media_type",
        416: "requested_range_not_satisfiable, requested_range, range_not_satisfiable",
        417: "expectation_failed",
        418: "im_a_teapot, teapot, i_am_a_teapot",
        421: "misdirected_request",
        422: "unprocessable_entity, unprocessable",
        423: "locked",
        424: "failed_dependency, dependency",
        425: "unordered_collection, unordered",
        426: "upgrade_required, upgrade",
        428: "precondition_required, precondition",
        429: "too_many_requests, too_many",
        431: "header_fields_too_large, fields_too_large",
        444: "no_response, none",
        449: "retry_with, retry",
        450: "blocked_by_windows_parental_controls, parental_controls",
        451: "unavailable_for_legal_reasons, legal_reasons",
        499: "client_closed_request",
        500: "internal_server_error, server_error, /o\, ✗",
        501: "not_implemented",
        502: "bad_gateway",
        503: "service_unavailable, unavailable",
        504: "gateway_timeout",
        505: "http_version_not_supported, http_version",
        506: "variant_also_negotiates",
        507: "insufficient_storage",
        509: "bandwidth_limit_exceeded, bandwidth",
        510: "not_extended",
        511: "network_authentication_required, network_auth, network_authentication",
    }


"""
This library allows you to quickly and easily use the Loyalicos Web API v1 via Python.
For more information on this library, see the README on GitHub.

For more information on the Loyalicos API, see:

For the user guide, code examples, and more, visit the main docs page:

This file provides the Loyalicos API Client.
"""

from .simplified_http_client import Response
import os
import requests
from .interface import Interface
from .exceptions import *
import json

class LoyalicosAPIClient(Interface):
    """ Loyaicos basic API client Object
        Extend these to add new objects with API interface.

        Use this object to interact with the Loyalicos API

    """

    def __init__(self, api_key=None, user_token={}, host=None, client_id=None, secret=None):
        self.user_token = user_token
        self.host = host 
        self.api_client = client_id 
        self.api_secret = secret 
        conf = {'LOYALICOS_API_CLIENT': None, 'LOYALICOS_API_SECRET': None, 'LOYALICOS_API_HOST':None, 'LOYALICOS_API_KEY':None }
        env = json.loads(os.environ.get('LOYALICOS_CONF'))
        conf.update(env)
        if self.api_client == None:
            self.api_client = conf.get('LOYALICOS_API_CLIENT')
        if self.api_secret == None:
            self.api_secret = conf.get('LOYALICOS_API_SECRET')
        if self.host == None:
            self.host = conf.get('LOYALICOS_API_HOST')
        self.api_key = api_key 
        if self.api_key == None:
            self.api_key = conf.get('LOYALICOS_API_KEY')
        if self.api_key != None:
            auth = 'Bearer {}'.format(self.api_key)
        else:
            if self.api_client == None or self.api_secret == None:
                raise NoCredentialsFoundError
            else:
                auth_response = requests.get(f'{self.host}/oauth/authapi', auth=requests.auth.HTTPBasicAuth(self.api_client, self.api_secret))
                if auth_response.status_code != 200:
                    raise NoCredentialsFoundError
                auth_result = auth_response.json()
                self.api_key = auth_result.get('token')
                auth = 'Bearer {}'.format(self.api_key)
        super(LoyalicosAPIClient, self).__init__(self.host, auth)

class Member(LoyalicosAPIClient):
    """
        Extends API Client to handle Members
    """
    def __init__(self, id='', api_key=None, user_token={}, host=None):
        self.id = id
        super(Member, self).__init__(api_key, user_token, host)

    """
        Add a Member
    """
    def create(self, alias=None, data={}):
        self.method = 'POST'
        self.path = ['member']
        self.json = {}
        self.json.update(data)
        self.json.update({"external_id" : alias})
        self.send_request()
        if self.response.status_code != 200:
            if self.response.status_code == 409:
                raise DuplicateKeyForMemberError
            raise HTTPRequestError
        self.access_token = self.response.body

    """
        Get a Member profile
    """
    def read(self, id=None, user_token={}):
        self.user_token.update(user_token)
        if id != None:
            self.id = id
        self.method = 'GET'
        self.path = ['member', self.id]
        user_auth = {'Access-Token' : self.user_token['access_token']}
        self.update_headers(user_auth)
        self.send_request()
        if self.response.status_code != 200:
            raise HTTPRequestError
        self.profile = self.response.body
        

    """
        Get a Member statement
    """
    def statement(self, user_token={}, filter_dict={}):
        filter_string = '|'.join([f"{v}={filter_dict[v]}" for v in filter_dict])
        if filter_dict != {}:
            self.params = f'filters={filter_string}'
        self.user_token.update(user_token)
        self.method = 'GET'
        self.path = ['member', 'awards', self.id]
        user_auth = {'Access-Token' : self.user_token['access_token']}
        self.update_headers(user_auth)
        self.send_request()
        if self.response.status_code != 200:
            raise HTTPRequestError
        self.stmt = self.response.body
        

    """
        Get a Member balance per type and code
    """
    def balance(self, type="currency", code="points", user_token={}):
        self.user_token.update(user_token)
        self.method = 'GET'
        self.path = ['member', 'profileValue', self.id, type, code]
        user_auth = {'Access-Token' : self.user_token['access_token']}
        self.update_headers(user_auth)
        self.send_request()
        if self.response.status_code != 200:
            raise HTTPRequestError
        self.profile = self.response.body
        

    """
        See if a member has an event type
    """
    def check_event(self,  id=None, user_token={}, search_params={}):
        self.user_token.update(user_token)
        if id != None:
            self.id = id
        self.method = 'POST'
        self.path = ['3PAMI', 'member', 'has_event', self.id]
        self.json = search_params
        user_auth = {'Access-Token' : self.user_token['access_token']}
        self.update_headers(user_auth)
        self.send_request()
        if self.response.status_code != 200:
            raise HTTPRequestError
        self.event_check = self.response.body.get('result')
        

    """
        Refresh a member token
    """
    def renew_token(self, user_token={}):
        self.user_token.update(user_token)
        self.method = 'POST'
        self.path = ['3PAMI', 'refreshToken']
        self.json = {'grant_type' : 'refresh_token', 'refresh_token' : user_token['refresh_token'] }
        self.send_request()
        if self.response.status_code != 200:
            if self.response.status_code == 404:
                raise MemberNotFoundError
            raise HTTPRequestError
        self.access_token = self.response.body
        

    """
        Get a new member token
    """
    def new_token(self, id = None):
        if id != None:
            self.id = id
        self.method = 'POST'
        self.path = ['3PAMI', 'accessToken']
        self.json = {'grant_type' : 'code', 'client_secret' : self.api_secret, 'external_id': self.id }
        self.send_request()
        if self.response.status_code != 200:
            raise HTTPRequestError
        self.access_token = self.response.body

class Event(LoyalicosAPIClient):
    """
        Extends API Client to handle events
    """
    def __init__(self, partner_code='',activity=None, type=None, channel=None,  member_id='', subactivity=None, subtype=None, subchannel=None, currency=None, date_activity=None, items=[], api_key=None, user_token={}, host=None):
        self.partner_code = partner_code
        self.external_id = member_id
        self.activity = activity
        self.type = type
        self.channel = channel
        self.subactivity = subactivity
        self.subtype = subtype
        self.subchannel = subchannel
        self.currency = currency
        self.date_activity = date_activity
        self.items = items
        super(Event, self).__init__(api_key, user_token, host)

    def make_body(self, format='json'):
        if format=='json':
            self.json = {
                'partner_code': self.partner_code,
                'external_id' : self.external_id,
                'date_activity' : self.date_activity,
                'channel': self.channel,
                'subchannel': self.subchannel,
                'type': self.type,
                'subtype': self.subtype,
                'activity': self.activity,
                'subactivity': self.subactivity,
                'currency': self.currency,
                'items': self.items
                }


    def loyalize(self, member_id='', date_activity=None):
        """
            Loyalize event
        """
        if member_id != '':
           self.external_id = member_id
        if date_activity != None:
           self.date_activity = date_activity
        self.method = 'PUT'
        self.path = ['loyalize']
        self.make_body('json')
        self.send_request()
        if self.response.status_code != 200:
            raise HTTPRequestError
        self.id = self.response.body['trx_id']

    def reward(self, member_id='', user_token={}, date_activity=None):
        """
            Send redemption transaction
        """
        if member_id != '':
           self.external_id = member_id
        if date_activity != None:
           self.date_activity = date_activity
        self.user_token.update(user_token)
        self.method = 'PUT'
        self.path = ['reward']
        self.make_body('json')
        user_auth = {'Access-Token' : self.user_token['access_token']}
        self.update_headers(user_auth)
        self.send_request()
        if self.response.status_code != 200:
            raise HTTPRequestError
        self.id = self.response.body['trx_id']